/*
 * FastaCheckerThreadPoolManager.java
 *
 * Created on Oct 19, 2007, 5:55:12 PM
 *
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package nl.uu.chem.bioinformatics.statquant.fastachecker;

import nl.uu.chem.bioinformatics.statquant.model.QuantifiedAminoAcidSequence;
import nl.uu.chem.bioinformatics.statquant.*;
import nl.uu.chem.bioinformatics.statquant.util.ThreadOptimum;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Observable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.uu.chem.bioinformatics.statquant.events.ProcessPreparingEvent;
import nl.uu.chem.bioinformatics.statquant.events.ProcessingFinishedEvent;
import nl.uu.chem.bioinformatics.statquant.events.ProgressEvent;
import nl.uu.chem.bioinformatics.statquant.model.FastaCheckerThreadOperation;
import nl.uu.chem.bioinformatics.statquant.model.FastaSequence;
import nl.uu.chem.bioinformatics.statquant.util.FastaParser;

/**
 *
 * @author Toorn101
 */
public class FastaCheckerThreadPool extends Observable implements Runnable {

    private final ProtsAndPepsDataHandler dataHandler;
    private Integer[] lengths;
    private HashMap<String, ArrayList<QuantifiedAminoAcidSequence>> map = new HashMap<String, ArrayList<QuantifiedAminoAcidSequence>>();
    private HashMap<String, ArrayList<QuantifiedAminoAcidSequence>> protsAndPeps = new HashMap<String, ArrayList<QuantifiedAminoAcidSequence>>();
    private FastaParser parser;
    private int numThreadsRunning = 0;
    private ConcurrentLinkedQueue<FastaCheckerThreadOperation> workqueue = new ConcurrentLinkedQueue<FastaCheckerThreadOperation>();

    public FastaCheckerThreadPool(ProtsAndPepsDataHandler dataHandler, File fastafile, String regex) throws FileNotFoundException {
        this.dataHandler = dataHandler;
        this.parser = new FastaParser(fastafile);
        this.parser.setTitlePattern(regex);
    }

    @Override
    public void run() {    
        /* Create a hash map of sequences, the values are
        lists of references back to the sequence data blocks */
        setChanged();
        notifyObservers(new ProcessPreparingEvent());
        protsAndPeps = dataHandler.getProtsAndPeps();
        for (String protein : protsAndPeps.keySet()) {
            for (QuantifiedAminoAcidSequence seq : protsAndPeps.get(protein)) {
                String ucSequence = seq.getSequence().toUpperCase();
                if (map.containsKey(ucSequence)) {
                    map.get(ucSequence).add(seq);
                } else {
                    ArrayList<QuantifiedAminoAcidSequence> list = new ArrayList<QuantifiedAminoAcidSequence>();
                    list.add(seq);
                    map.put(ucSequence, list);
                }
            }
        }


        /* Create a list with the lengths of known peptides,
        as only these should be searched */
        HashSet<Integer> lengthsset = new HashSet<Integer>();
        for (String key : map.keySet()) {
            lengthsset.add(key.length());
        }

        lengths = lengthsset.toArray(new Integer[0]);

        try {
            int numthreads = ThreadOptimum.measure();
            HashMap<String, Void> placeholder = new HashMap<String, Void>();

            for (String key : map.keySet()) {
                placeholder.put(key, null);
            }

            for (int a = 0; a < numthreads; a++) {
                @SuppressWarnings(value = "unchecked")
                FastaCheckerThread checker = new FastaCheckerThread(this, (HashMap<String, Void>) placeholder.clone(), a);                
                checker.start();
                numThreadsRunning++;
            }
        } catch (InterruptedException ignored) {
        }
    }

    synchronized FastaCheckerThreadOperation getNewAssignment() {
        FastaSequence currentSequence;
        if (workqueue.isEmpty()) {
            try {
                currentSequence = parser.getSequence();
                setChanged();
                notifyObservers(new ProgressEvent(parser.getFileProgress()));
                if (currentSequence != null) {
                    for (int length : lengths) {
                        workqueue.add(new FastaCheckerThreadOperation(currentSequence, length));
                    }
                }
            } catch (IOException ex) {
                Logger.getLogger(FastaCheckerThreadPool.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        if (workqueue.isEmpty()) {
            return null;
        } else {
            return workqueue.poll();
        }
    }

    synchronized void doneWorking() {
        if (--numThreadsRunning == 0) {
            try {
                int numfound = 0;
                //Second pass is to set the selected flag
                for (String protein : protsAndPeps.keySet()) {
                    for (QuantifiedAminoAcidSequence seq : protsAndPeps.get(protein)) {
                        if (!seq.isPeptideUnique()) {
                          //  seq.setChecked(false);
                            numfound++;
                        }
                    }
                }

                dataHandler.setFilteredForUniqueness(true);
                parser.close();
                setChanged();
                notifyObservers(new ProcessingFinishedEvent("Found " + numfound + " non unique peptides"));
            } catch (IOException ex) {
                Logger.getLogger(FastaCheckerThreadPool.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    synchronized void updatePeptide(String sequence, String identifier) {

        for (QuantifiedAminoAcidSequence quantseq : map.get(sequence)) {
            quantseq.addToProteinList(identifier);
        }
    }
}